<?php
/*
Plugin Name: Post Revision Display
Plugin URI: http://www.movingtofreedom.org/10/07/30/wordpress-plugin-post-revision-and-diff-viewer
Description: Show list of post-publication revisions on single post pages, and provide links to display them along with diffs between old and current revision. Revision pages are marked "noindex" for robot crawlers. Inspired by <a href="http://www.wordyard.com/2010/07/23/help-with-a-wordpress-plugin-for-published-versions/">Scott Rosenberg's request for a plugin to demo versioning on blog posts</a> and built on <a href="http://www.darcynorman.net/wordpress/post-revision-display/">D'Arcy Norman's original plugin</a>.
Version: 0.6
Author: Scott Carpenter, D'Arcy Norman
Author URI: http://www.movingtofreedom.org
License: GPL2
*/
/*  Copyright 2008  D'Arcy Norman  (email : dlnorman@ucalgary.ca)
    Copyright 2010  Scott Carpenter (email: scottc@movingtofreedom.org)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

define('REV_LIST_HEADER', '<h4>Post Revisions:</h4>');
define('REV_DIFFS_HEADER', '<h4>Changes:</h4>');

add_action('wp_head', 'prd_noindex_header');
add_filter('the_content', 'prd_display_post_revisions');

function prd_noindex_header()
{
	if (isset($_GET['rev'])) {
		// like Wikipedia does for its old revisions
		echo "\t\t" . '<meta name="robots" content="noindex, nofollow" />' . "\n";
	}
}

function prd_display_post_revisions($content) {

	if (!$post = get_post(get_the_ID())) {
		return $content;
	}
	if (empty($GLOBALS['prd_mode'])) {
		// "automatic" mode -- everything is handled here
		if (!is_single()) {
			return $content;
		}
		prd_set_globals();
		if ($GLOBALS['prd_is_rev']) {
			return $GLOBALS['prd_revision_note'] . $GLOBALS['prd_revision_content'] .
			       $GLOBALS['prd_revision_list'] . $GLOBALS['prd_revision_diffs'];
		} else {
			return $content . $GLOBALS['prd_revision_list'];
		}
	} else {
		// "manual" mode -- calls are made from theme to specify where things go
		if ($GLOBALS['prd_is_rev']) {
			return $GLOBALS['prd_revision_content'];
		} else {
			return $content;
		}
	}
}

// question: should we be applying filters to all the parts and not just content?
function prd_set_globals() {
	if (!$post = get_post(get_the_ID())) {
		return;
	}
	$revision = null;
	$rev_id = 0;
	$is_rev = false;
	$note = prd_get_revision_note($post, $revision, $rev_id, $is_rev);
	if ($is_rev) {
		$diffs = prd_get_revision_diffs($post, $revision);
		
		/* seems clunky, but I don't know a better way -- want to apply_filters
		   for old revision, but need to skip prd_display_post_revisions filter
		   to avoid getting stuck in a loop, so will remove filter and add
		   right back in */
		remove_filter('the_content', 'prd_display_post_revisions');
		$rev_content = apply_filters('the_content', $revision->post_content);
		add_filter('the_content', 'prd_display_post_revisions');
	}
	$revs = prd_get_revision_list($post, array('since_publish' => true, 'type' => 'revision',
	                                           'rev_id' => $rev_id, 'is_rev' => $is_rev));
	$GLOBALS['prd_is_rev'] = $is_rev;
	$GLOBALS['prd_revision_note'] = $note;
	$GLOBALS['prd_revision_list'] = $revs;
	$GLOBALS['prd_revision_diffs'] = $diffs;
	$GLOBALS['prd_revision_content'] = $rev_content;
}

// kind of like wp_list_post_revisions()
function prd_get_revision_list($post, $args=null)
{
	$defaults = array( 'parent' => false, 'right' => false, 'left' => false, 'format' => 'list', 'type' => 'all', 'since_publish' => false, 'rev_id' => 0, 'is_rev' => false );
	extract( wp_parse_args( $args, $defaults ), EXTR_SKIP );

	$rev_list = '';
	switch ($type) {
		case 'autosave' :
			if (!$autosave = wp_get_post_autosave($post->ID)) {
				$rev_list = "<p>No autosave</p>";
			}
			$revisions = array( $autosave );
			break;
		case 'revision' : // just revisions - remove autosave later
		case 'all' :
		default :
			if (!$revisions = wp_get_post_revisions($post->ID)) {
				$rev_list = "<p>There are no revisions for this post.</p>";
			}
			break;
	}

	$titlef = _c('%1$s by %2$s|post revision 1:datetime, 2:name');

	if ($parent) {
		array_unshift( $revisions, $post );
	}

	$rows = '';
	$class = false;
	//$can_edit_post = current_user_can( 'edit_post', $post->ID ); // not used?
	if ($since_publish) {
		$post_time = strtotime($post->post_date_gmt);
	}
	foreach ($revisions as $revision) {
		if ('revision' === $type && wp_is_post_autosave($revision)) {
			continue;
		}
		if ($since_publish && strtotime($revision->post_date_gmt) < $post_time) {
			continue;
		}
		// 2nd param to wp_post_revision_title determines if link (if true, only links if you have access)
		$date = wp_post_revision_title($revision, false);
		if ($rev_id != $revision->ID) {
			$date = '<a href="' . get_permalink() . '?rev=' . $revision->ID . '">' . "$date</a>";
		}
		$name = get_author_name($revision->post_author);

		$title = sprintf( $titlef, $date, $name );
		$rows .= "\t<li>$title</li>\n";
	}

	if (!empty($rows)) {
		// add current revision to the top
		$date = wp_post_revision_title($post, false);
		if ($is_rev) {
			$date = '<a href="' . get_permalink() . '">' . $date . '</a>';
		}
		$name = get_author_name($post->post_author);
		$title = sprintf($titlef, $date, $name);
		$rows = "\t<li>$title</li>\n$rows";

		$rev_list = '<ul class="post-revisions">' .
					"\n$rows\n</ul>";
	} else if (empty($rev_list)) {
		$rev_list = "<p>This post has not been revised since publication.</p>";
	}
	return '<div class="post-revisions">' . "\n" .
		   REV_LIST_HEADER .
		   "\n$rev_list\n</div>\n";
}

// only passing $rev_id and $is_rev here so that we can use them in
// prd_get_revision_list() to check and not link the displayed revision --
// is it worth the trouble ???
function prd_get_revision_note($post, &$revision, &$rev_id, &$is_rev)
{
	$is_rev = false;
	$note = '';
	if (isset($_GET['rev'])) {
		$rev_id = intval($_GET['rev']);
		$current_rev_link = '<a href="' . get_permalink() . '">current revision</a>';
		$view_current = "(Viewing $current_rev_link instead.)";
		if (!$revision = get_post($rev_id)) {
			$note = "Revision $rev_id not found. $view_current";
		} else {
			$post_id = $post->ID;
			if ($revision->post_parent == 0) {
				$note = 'This is the current revision.';
			} else if ($revision->post_parent != $post->ID) { // this check should come before the date check
				$note = "Revision $rev_id is not a revision of this post. $view_current";
			} else if (strtotime($revision->post_date_gmt) < strtotime($post->post_date_gmt)) {
				$note = "Revision $rev_id is a pre-publication revision. $view_current";
			} else {
				$is_rev = true;
				$note = 'You are viewing an old revision of this post, from ' . wp_post_revision_title($revision, false) . '. <a href="#revision-diffs">See below for differences</a> between this version and the ' . "$current_rev_link.";
			}
		}
		$note = '<div class="revision-header"><p>' . $note . "</p></div>\n";
	}
	return $note;
}

// TODO: use code from revision.php and pluggable.php to get a nice side-by-side and color-coded diff view
function prd_get_revision_diffs($post, $revision)
{
	//	$diffs = wp_text_diff($revision->post_content, $post->post_content); // built-in call

	// diff the unfiltered content
	$diffs = PHPDiff(trim($revision->post_content), trim($post->post_content));
	// check post title for changes
	if (strcmp(trim($revision->post_title), trim($post->post_title)) != 0) {
		$diffs = "title:\n" .
				 '  <span style="color: red;">- ' . $revision->post_title . "</span>\n" .
				 '  <span style="color: blue;">+ ' . $post->post_title . "</span>\n" .
				$diffs;
	}
	if (empty($diffs)) {
		$diffs = '<p>There are no differences between the ' . wp_post_revision_title($revision, false) .
		         ' revision and the current revision. (Maybe only post meta information was changed.)</p>';
	} else {
		$diffs = '<p>Differences between ' . wp_post_revision_title($revision, false) . ' revision (<span style="color: red;">- red</span>) and current revision (<span style="color: blue;">+ blue</span>). Line numbers on left of [diff type] are for the old revision and on right are for the current.</p>' . "\n" . '<pre style="overflow: auto;">'. "$diffs</pre>";
	}
	return '<div id="revision-diffs">' . "\n" .
	       REV_DIFFS_HEADER .
	       "\n$diffs\n</div>\n";
}

/*
 * Manual Mode
 *
 * prd_set_manual_mode only needs to be called if you're wanting to use 
 * manual mode and one of the_revision_* functions is not called
 * *before* the_content (if nothing is called until after the_content,
 * then our filter will assume "automatic" mode and do stuff, um, automatically
 *
 */
function prd_set_manual_mode() {
	prd_set_globals();
	$GLOBALS['prd_mode'] = 'manual';	
}

function the_revision_note_prd() {
	if (empty($GLOBALS['prd_mode'])) {
		prd_set_globals();
	}
	$GLOBALS['prd_mode'] = 'manual';
	echo $GLOBALS['prd_revision_note'];
}

// rev list is the only thing that really makes sense to show in the loop --
// can set $refreshGlobals to true to force it to update for each post
// $header is if you want something other than <h4>Revision List:</h4>
function the_revision_list_prd($refreshGlobals=false, $header=REV_LIST_HEADER) {
	if (empty($GLOBALS['prd_mode']) || $refreshGlobals) {
		prd_set_globals();
	}
	if (strcmp($header, REV_LIST_HEADER) != 0) {
		$GLOBALS['prd_revision_list'] =
			str_replace(REV_LIST_HEADER, $header, $GLOBALS['prd_revision_list']);
	}
	$GLOBALS['prd_mode'] = 'manual';
	echo $GLOBALS['prd_revision_list'];
}

// $header is if you want something other than <h4>Changes:</h4>
function the_revision_diffs_prd($header=REV_DIFFS_HEADER) {
	if (empty($GLOBALS['prd_mode'])) {
		prd_set_globals();
	}
	if (strcmp($header, REV_DIFFS_HEADER) != 0) {
		$GLOBALS['prd_revision_diffs'] =
			str_replace(REV_DIFFS_HEADER, $header, $GLOBALS['prd_revision_diffs']);
	}
	$GLOBALS['prd_mode'] = 'manual';
	echo $GLOBALS['prd_revision_diffs'];
}

/**
    Diff implemented in pure php, written from scratch.
    Copyright (C) 2003  Daniel Unterberger <diff.phpnet@holomind.de>
    Copyright (C) 2005  Nils Knappmeier next version

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

    http://www.gnu.org/licenses/gpl.html

    About:
    I searched a function to compare arrays and the array_diff()
    was not specific enough. It ignores the order of the array-values.
    So I reimplemented the diff-function which is found on unix-systems
    but this you can use directly in your code and adopt for your needs.
    Simply adopt the formatline-function. with the third-parameter of arr_diff()
    you can hide matching lines. Hope someone has use for this.

    Contact: d.u.diff@holomind.de <daniel unterberger>
**/

// original code pages:
// http://www.holomind.de/phpnet/diff.src.php
// http://www.holomind.de/phpnet/diff2.src.php

## PHPDiff returns the differences between $old and $new, formatted
## in the standard diff(1) output format.
## 28 July 2010 -- modified here to be non-standard!
function PHPDiff($old, $new)
{
   # split the source text into arrays of lines
   $t1 = explode("\n",$old);
   $x=array_pop($t1);
   //if ($x>'') $t1[]="$x\n\\ No newline at end of file";
   if ($x>'') $t1[]="$x\n";
   $t2 = explode("\n",$new);
   $x=array_pop($t2);
   //if ($x>'') $t2[]="$x\n\\ No newline at end of file";
   if ($x>'') $t2[]="$x\n";

   # build a reverse-index array using the line as key and line number as value
   # don't store blank lines, so they won't be targets of the shortest distance
   # search
   foreach($t1 as $i=>$x) if ($x>'') $r1[$x][]=$i;
   foreach($t2 as $i=>$x) if ($x>'') $r2[$x][]=$i;

   $a1=0; $a2=0;   # start at beginning of each list
   $actions=array();

   # walk this loop until we reach the end of one of the lists
   while ($a1<count($t1) && $a2<count($t2)) {
     # if we have a common element, save it and go to the next
     if ($t1[$a1]==$t2[$a2]) { $actions[]=4; $a1++; $a2++; continue; }

     # otherwise, find the shortest move (Manhattan-distance) from the
     # current location
     $best1=count($t1); $best2=count($t2);
     $s1=$a1; $s2=$a2;
     while(($s1+$s2-$a1-$a2) < ($best1+$best2-$a1-$a2)) {
       $d=-1;
       foreach((array)@$r1[$t2[$s2]] as $n)
         if ($n>=$s1) { $d=$n; break; }
       if ($d>=$s1 && ($d+$s2-$a1-$a2)<($best1+$best2-$a1-$a2))
         { $best1=$d; $best2=$s2; }
       $d=-1;
       foreach((array)@$r2[$t1[$s1]] as $n)
         if ($n>=$s2) { $d=$n; break; }
       if ($d>=$s2 && ($s1+$d-$a1-$a2)<($best1+$best2-$a1-$a2))
         { $best1=$s1; $best2=$d; }
       $s1++; $s2++;
     }
     while ($a1<$best1) { $actions[]=1; $a1++; }  # deleted elements
     while ($a2<$best2) { $actions[]=2; $a2++; }  # added elements
  }

  # we've reached the end of one list, now walk to the end of the other
  while($a1<count($t1)) { $actions[]=1; $a1++; }  # deleted elements
  while($a2<count($t2)) { $actions[]=2; $a2++; }  # added elements

  # and this marks our ending point
  $actions[]=8;

  # now, let's follow the path we just took and report the added/deleted
  # elements into $out.
  $op = 0;
  $x0=$x1=0; $y0=$y1=0;
  $out = array();
  foreach($actions as $act) {
    if ($act==1) { $op|=$act; $x1++; continue; }
    if ($act==2) { $op|=$act; $y1++; continue; }
    if ($op>0) {
      $xstr = ($x1==($x0+1)) ? $x1 : ($x0+1).",$x1";
      $ystr = ($y1==($y0+1)) ? $y1 : ($y0+1).",$y1";
      if ($op==1) $out[] = "lines: {$xstr} [delete] {$y1}";
      elseif ($op==3) $out[] = "lines: {$xstr} [change] {$ystr}";
      while ($x0<$x1) { $out[] = '&nbsp;&nbsp;<span style="color: red;">- ' .
                                 htmlspecialchars(trim($t1[$x0])) . '</span>'; $x0++; }   # deleted elems
      if ($op==2) $out[] = "lines: {$x1} [add] {$ystr}";
      //elseif ($op==3) $out[] = '---';
      while ($y0<$y1) { $out[] = '&nbsp;&nbsp;<span style="color: blue;">+ ' .
                                 htmlspecialchars(trim($t2[$y0])) . '</span>'; $y0++; }   # added elems
    }
    $x1++; $x0=$x1;
    $y1++; $y0=$y1;
    $op=0;
  }
  $out[] = '';
  return join("\n",$out);
}
?>
